//
//  ContentView.swift
//  Shared
//
//  Created by Jeff Campbell on 1/12/21.
//

import SwiftUI

struct ContentView: View {
	enum Section {
		case access
		case error
	}

	@State private var section:Section			= .access
	
	@State private var urlString				= "https://wbur-sgrewind.streamguys1.com/wbur/wbur/chunks_dvr.m3u8"

	@ObservedObject private var streamManager	= StreamManager()

	var body: some View {
		VStack {
			Text("HLS Stream Tester")
				.padding()

			TextField("Stream URL", text: self.$urlString)
				.frame(minWidth: 200, idealWidth: 320, maxWidth: .infinity)
				.textFieldStyle(RoundedBorderTextFieldStyle())
				.padding(.horizontal)

			Button(action: {
				self.streamManager.pause()

				self.streamManager.configure(withURL: URL(string: self.urlString.trimmingCharacters(in: .whitespaces))!)
			}, label: {
				Text("Configure")
			})
			.padding(.horizontal)
			.frame(minWidth: 100)

			HStack {
				Button(action: {
					self.streamManager.skipBackward(30) {
					}
				}, label: {
					Text("◀ 30")
				})
				.padding(.horizontal)
				.disabled(!self.streamManager.state.isPlaying)
				.frame(minWidth: 100)

				if self.streamManager.state.isReady == true {
					Button(action: {
						if self.streamManager.state.isPlaying == false {
							self.streamManager.play()
						} else {
							self.streamManager.pause()
						}
					}, label: {
						if self.streamManager.state.isPlaying == true {
							Text("Pause")
						} else {
							Text("Play")
						}
					})
					.padding(.horizontal)
					.frame(minWidth: 100)
				} else {
					Text("-")
				}

				Button(action: {
					self.streamManager.skipForward(30) {
					}
				}, label: {
					Text("30 ▶")
				})
				.padding(.horizontal)
				.disabled(!self.streamManager.state.isPlaying)
				.frame(minWidth: 100)
			}
			.padding(.bottom)
			.padding(.horizontal)

			if let state = self.streamManager.state {
				VStack {
					Text("\(Int(state.currentTime.seconds.rounded(.down))) of \(Int(self.streamManager.loadedTimeRange().end.seconds.rounded(.down))) (Δ \(Int(state.currentTime.seconds.rounded(.down)) - Int(self.streamManager.loadedTimeRange().end.seconds.rounded(.down))))" as String)
						.padding(.horizontal)
					if let currentDate = state.currentDate {
						Text("\(currentDate)")
							.padding(.horizontal)
					}
				}
			} else {
				Text(" ")
					.padding(.horizontal)
			}

			Picker(selection: self.$section, label: EmptyView()) {
				Text("Access").tag(Section.access)
				Text("Error").tag(Section.error)
			}.pickerStyle(SegmentedPickerStyle())
			.padding()

			ScrollView {
				VStack {
					if let state = self.streamManager.state {
//						Text("Duration: \(Int(state.duration.seconds.rounded(.down)))" as String)
//						Text("Current Time: \(Int(state.currentTime.seconds.rounded(.down)))" as String)
//						Text("Forward Playback End Time: \(state.forwardPlaybackEndTime.seconds.rounded(.down))" as String)
//
//						Text("Loaded Time Ranges: \(state.loadedTimeRanges.count)" as String)
//						Text("Seekable Time Ranges: \(state.seekableTimeRanges.count)" as String)
//
//						if let lastError = state.lastError, let lastErrorDate = state.lastErrorDate {
//							Text("Error: \(lastError.localizedDescription) (\(lastErrorDate)")
//						}

						if self.section == .access {
							if let playerItemAccessLogData = state.playerItemAccessLog?.extendedLogData(), let playerItemAccessLogString = String(data: playerItemAccessLogData, encoding: .utf8) {
								Text(playerItemAccessLogString)
									.frame(maxWidth: .infinity, alignment: .leading)
									.multilineTextAlignment(.leading)
									.contextMenu {
											Button(action: {
												#if os(iOS)
												UIPasteboard.general.string = playerItemAccessLogString
												#elseif os(macOS)
												NSPasteboard.general.declareTypes([.string], owner: nil)
												NSPasteboard.general.setString(playerItemAccessLogString, forType: .string)
												#endif
											}) {
												Text("Copy")
												Image(systemName: "doc.on.doc")
											}
										 }
							}
						} else {
							if let playerItemErrorLogData = state.playerItemErrorLog?.extendedLogData(), let playerItemErrorLogString = String(data: playerItemErrorLogData, encoding: .utf8) {
								Text(playerItemErrorLogString)
									.frame(maxWidth: .infinity, alignment: .leading)
									.multilineTextAlignment(.leading)
									.contextMenu {
											Button(action: {
												#if os(iOS)
												UIPasteboard.general.string = playerItemErrorLogString
												#elseif os(macOS)
												NSPasteboard.general.declareTypes([.string], owner: nil)
												NSPasteboard.general.setString(playerItemErrorLogString, forType: .string)
												#endif
											}) {
												Text("Copy")
												Image(systemName: "doc.on.doc")
											}
										 }
							}
						}
					}
				}
				.frame(maxWidth: .infinity)
				.padding()
			}
			.frame(maxWidth: .infinity)
			.overlay(Rectangle().frame(width: nil, height: 1, alignment: .top).foregroundColor(Color.gray), alignment: .top)
			.onTapGesture {
				#if os(iOS)
				UIApplication.shared.sendAction(#selector(UIResponder.resignFirstResponder), to: nil, from: nil, for: nil)
				#endif
			}
		}
		.frame(minWidth: 300, minHeight: 500)
	}
}

struct ContentView_Previews: PreviewProvider {
    static var previews: some View {
        ContentView()
    }
}
