//
//  StreamManager.swift
//
//  Created by Jeff Campbell on 1/12/21.
//

import Foundation
import AVFoundation
import MediaPlayer
import SwiftUI

class StreamManager: NSObject, ObservableObject {
	@objc public internal(set) var audioPlayer:AVPlayer?

	@Published public internal(set) var isLoading	= false

	@Published public internal(set) var state		= PlayerState(duration: CMTime.zero, currentTime: CMTime.zero, currentDate: nil, seekableTimeRanges: [], loadedTimeRanges: [], forwardPlaybackEndTime: CMTime.zero, playerItemAccessLog: nil, playerItemErrorLog: nil, isReady: false, isPlaying: false, lastError: nil, lastErrorDate: nil)

	private var playerCurrentItemToken: NSKeyValueObservation?
	private var playerRateToken: NSKeyValueObservation?

	private var playerPeriodicTimeToken: Any?

	private var playerItemDurationToken: NSKeyValueObservation?
	private var playerItemStatusToken: NSKeyValueObservation?
	private var playerItemLoadedTimeRangesToken: NSKeyValueObservation?
	private var playerItemSeekableTimeRangesToken: NSKeyValueObservation?

	public override init() {
		super.init()
	}

	func updateState() {
		guard let audioPlayer = self.audioPlayer, let currentItem = audioPlayer.currentItem else { return }

		var isReady = false
		if audioPlayer.currentItem != nil {
			isReady = true
		}

		var state = PlayerState(duration: currentItem.duration, currentTime: audioPlayer.currentTime(), seekableTimeRanges: currentItem.seekableTimeRanges, loadedTimeRanges: currentItem.loadedTimeRanges, forwardPlaybackEndTime: currentItem.forwardPlaybackEndTime, playerItemAccessLog:currentItem.accessLog(), playerItemErrorLog:currentItem.errorLog(), isReady: isReady, isPlaying: self.audioPlayer?.rate != 0.0)

		state.currentDate = audioPlayer.currentItem?.currentDate()

		if let error = audioPlayer.currentItem?.error {
			if let lastError = state.lastError {
				if lastError.localizedDescription != error.localizedDescription {
					state.lastError		= error
					state.lastErrorDate	= Date()
				}
			} else {
				state.lastError		= error
				state.lastErrorDate	= Date()
			}
		} else if let error = audioPlayer.error {
			if let lastError = state.lastError {
				if lastError.localizedDescription != error.localizedDescription {
					state.lastError		= error
					state.lastErrorDate	= Date()
				}
			} else {
				state.lastError		= error
				state.lastErrorDate	= Date()
			}
		}

		DispatchQueue.main.async {
			self.state	= state

			self.updateNowPlaying(withState: state)
			self.updateRemoteCommands(withState: state)

			print("isPlaying: \(state.isPlaying), loadedTimeRanges: \(state.loadedTimeRanges as Any)")
		}
	}

	func configure(withURL url:URL) {
		self.audioPlayer?.pause()

		UIApplication.shared.beginReceivingRemoteControlEvents()

		let asset				= AVURLAsset(url: url)

		asset.loadValuesAsynchronously(forKeys: ["duration"], completionHandler: {
		})

		let playerItem			= AVPlayerItem(asset: asset)

//		print("playerItem.automaticallyPreservesTimeOffsetFromLive = \(playerItem.automaticallyPreservesTimeOffsetFromLive)")
//		print("playerItem.canUseNetworkResourcesForLiveStreamingWhilePaused = \(playerItem.canUseNetworkResourcesForLiveStreamingWhilePaused)")
//		print("playerItem.preferredForwardBufferDuration = \(playerItem.preferredForwardBufferDuration)")

		#if os(iOS)
		try? AVAudioSession.sharedInstance().setCategory(AVAudioSession.Category.playback, mode: AVAudioSession.Mode.default, options: [])
		#endif

		self.audioPlayer		= AVPlayer(playerItem: playerItem)

		if let audioPlayer = self.audioPlayer {
			// Player Observers...

			self.playerCurrentItemToken = audioPlayer.observe(\.currentItem) { player, _ in
				self.updateState()
			}

			self.playerRateToken = audioPlayer.observe(\.rate) { player, _ in
				self.updateState()
			}

			self.playerPeriodicTimeToken = audioPlayer.addPeriodicTimeObserver(forInterval: CMTimeMakeWithSeconds(1.0, preferredTimescale: 100000), queue: DispatchQueue.global(qos: .default), using: { (time) in
				self.updateState()
			})

			// Player Item Observers...

			self.playerItemDurationToken = playerItem.observe(\.duration) { playerItem, _ in
				self.updateState()
			}

			self.playerItemStatusToken = playerItem.observe(\.status) { playerItem, _ in
				self.updateState()
			}

			self.playerItemSeekableTimeRangesToken = playerItem.observe(\.seekableTimeRanges) { playerItem, _ in
				self.updateState()
			}

			self.playerItemLoadedTimeRangesToken = playerItem.observe(\.loadedTimeRanges) { playerItem, _ in
				self.updateState()
			}

			audioPlayer.play()
		}
	}

	func play() {
		try? AVAudioSession.sharedInstance().setActive(true)

		self.audioPlayer?.play()

		self.updateState()
	}

	func pause() {
		self.audioPlayer?.pause()

		self.updateState()
	}

	func skipBackward(_ seconds:Double, completion: (() -> Void)? = nil) {
		self.updateState()

		DispatchQueue.global(qos: .userInteractive).async {
			guard let currentTimeSeconds = self.audioPlayer?.currentTime().seconds else { return }

			self.seek(toSeconds: currentTimeSeconds - seconds, currentItemOnly: true, completion: completion)

			self.updateState()
		}
	}

	func skipForward(_ seconds:Double, completion: (() -> Void)? = nil) {
		self.updateState()

		DispatchQueue.global(qos: .userInteractive).async {
			guard let currentTimeSeconds = self.audioPlayer?.currentTime().seconds else { return }

			self.seek(toSeconds: currentTimeSeconds + seconds, currentItemOnly: true, completion: completion)

			self.updateState()
		}
	}

	func seek(toSeconds seconds:Double, currentItemOnly:Bool = false, completion: (() -> Void)? = nil) {
		self.updateState()

		DispatchQueue.global(qos: .userInteractive).async {
			// Determine seek time... - JAC
			let normalizedSeconds			= seconds.rounded(.down)

			// Seek to seek time... - JAC
			let seekTime = CMTimeMakeWithSeconds(normalizedSeconds, preferredTimescale: 100000)

			self.seek(toTime: seekTime, completion: {
				self.updateState()

				completion?()
			})
		}
	}

	func seek(toTime seekTime:CMTime, completion: (() -> Void)?) {
		guard let audioPlayer = self.audioPlayer else { return }

		audioPlayer.currentItem?.cancelPendingSeeks()

		audioPlayer.seek(to: seekTime, toleranceBefore: CMTime.zero, toleranceAfter: CMTime.zero) { (finished) in
			self.updateState()

			if finished == true {
				completion?()
			}
		}
	}

	func loadedTimeRange() -> CMTimeRange {
		if let loadedTimeRange = self.audioPlayer?.currentItem!.loadedTimeRanges.last?.timeRangeValue {
			return loadedTimeRange
		}

		return CMTimeRange(start: CMTime.zero, end: CMTime.zero)
	}

	func updateNowPlaying(withState state:PlayerState) {
		var nowPlayingInfoDictionary:[String:Any]									= [:]

		nowPlayingInfoDictionary[MPNowPlayingInfoPropertyIsLiveStream]				= NSNumber(booleanLiteral: true)

		if audioPlayer?.rate == 1.0 {
			nowPlayingInfoDictionary[MPNowPlayingInfoPropertyPlaybackRate]			= NSNumber(value: 1.0)
		} else {
			nowPlayingInfoDictionary[MPNowPlayingInfoPropertyPlaybackRate]			= NSNumber(value: 0.0)
		}

		nowPlayingInfoDictionary[MPMediaItemPropertyTitle]							= "Test Stream"

		// Misc...
		nowPlayingInfoDictionary[MPNowPlayingInfoPropertyMediaType]					= NSNumber(value: MPNowPlayingInfoMediaType.audio.rawValue)

		MPNowPlayingInfoCenter.default().nowPlayingInfo								= nowPlayingInfoDictionary
	}

	func updateRemoteCommands(withState state:PlayerState) {
		MPRemoteCommandCenter.shared().skipBackwardCommand.preferredIntervals		= [NSNumber(value: 30)]
		MPRemoteCommandCenter.shared().skipForwardCommand.preferredIntervals		= [NSNumber(value: 30)]

		MPRemoteCommandCenter.shared().skipBackwardCommand.isEnabled 				= true
		MPRemoteCommandCenter.shared().skipForwardCommand.isEnabled 				= true

		MPRemoteCommandCenter.shared().pauseCommand.isEnabled						= true
		MPRemoteCommandCenter.shared().playCommand.isEnabled						= true

		MPRemoteCommandCenter.shared().togglePlayPauseCommand.addTarget { (event) -> MPRemoteCommandHandlerStatus in
			if self.state.isPlaying {
				self.pause()
			} else {
				self.play()
			}

			return .success
		}

		MPRemoteCommandCenter.shared().pauseCommand.addTarget(self, action: #selector(self.handlePauseRemote(_:)))
		MPRemoteCommandCenter.shared().playCommand.addTarget(self, action: #selector(self.handlePlayRemote(_:)))
		MPRemoteCommandCenter.shared().skipBackwardCommand.addTarget(self, action: #selector(self.handleSkipBackwardRemote(_:)))
		MPRemoteCommandCenter.shared().skipForwardCommand.addTarget(self, action: #selector(self.handleSkipForwardRemote(_:)))
	}

	@objc func handlePauseRemote(_ event: MPRemoteCommandEvent) -> MPRemoteCommandHandlerStatus {
		self.pause()

		return .success
	}

	@objc func handlePlayRemote(_ event: MPRemoteCommandEvent) -> MPRemoteCommandHandlerStatus {
		self.play()

		return .success
	}

	@objc func handleSkipBackwardRemote(_ event: MPRemoteCommandEvent) -> MPRemoteCommandHandlerStatus {
		self.skipBackward(30) {
			print("Skipped backward...")

			self.updateState()
		}

		return .success
	}

	@objc func handleSkipForwardRemote(_ event: MPRemoteCommandEvent) -> MPRemoteCommandHandlerStatus {
		self.skipForward(30) {
			print("Skipped forward...")

			self.updateState()
		}

		return .success
	}
}
